import RPi.GPIO
from gfxlcd.abstract.touch import Touch


class XPT2046(Touch):
    """XPT2046 class"""
    def __init__(self, width, height, int_pin=None, callback=None, cs_pin=None, spi=0, speed=1000000):
        super().__init__(width, height, int_pin, callback, cs_pin, spi, speed)
        self.correction = {
            'x': 540,
            'y': 50,
            'ratio_x': 0.94,
            'ratio_y': 1.26,
        }

    def _get_xy(self, offset_x, offset_y):
        """correct x and y"""
        if self.rotate == 0:
            return int((offset_x - self.correction['x']) / self.correction['ratio_x']), \
                self.height - int((offset_y - self.correction['y']) / self.correction['ratio_y'])

        if self.rotate == 90:
            return int((offset_y - self.correction['y']) / self.correction['ratio_y']), \
                int((offset_x - self.correction['x']) / self.correction['ratio_x']),

        if self.rotate == 180:
            return self.width - int((offset_x - self.correction['x']) / self.correction['ratio_x']), \
                int((offset_y - self.correction['y']) / self.correction['ratio_y'])

        if self.rotate == 270:
            return self.height - int((offset_y - self.correction['y']) / self.correction['ratio_y']), \
                self.width - int((offset_x - self.correction['x']) / self.correction['ratio_x'])

    def get_position(self):
        """get touch coords"""
        buffer = []
        fuse = 40
        while len(buffer) < 20 and fuse > 0:
            if self.cs_pin:
                RPi.GPIO.output(self.cs_pin, 0)

            self.spi.xfer2([0x80 | 0x08 | 0x30])
            recv = self.spi.readbytes(1)
            tc_rz = recv[0] & 0x7f

            self.spi.xfer2([0x80 | 0x08 | 0x40])
            recv = self.spi.readbytes(1)
            tc_rz += (255-recv[0] & 0x7f)

            self.spi.xfer2([0x80 | 0x10])
            recv = self.spi.readbytes(2)
            tc_rx = 1023-((recv[0] << 2)|(recv[1] >> 6))

            self.spi.xfer2([0x80 | 0x50])
            recv = self.spi.readbytes(2)
            tc_ry = ((recv[0] << 2)|(recv[1] >> 6))

            if self.cs_pin:
                RPi.GPIO.output(self.cs_pin, 1)
            if tc_rz > 10:
                pos_x, pos_y = self._get_xy(tc_rx, tc_ry)
                if self._in_bounds(pos_x, pos_y):
                    buffer.append((pos_x, pos_y))
            fuse -= 1

        return self._calculate_avr(buffer)



"""Null Area test chip driver"""
from gfxlcd.drawing.area import Area
from gfxlcd.abstract.chip import Chip


class NullArea(Area, Chip):
    """Test chip driver for area drawing"""
    def __init__(self, width, height, driver, auto_flush=True):
        Chip.__init__(self, width, height, driver, auto_flush)
        Area.__init__(self, driver)
        self.rotation = 0

    def _converted_background_color(self):
        """dummy background color"""
        return 2

    def _converted_color(self):
        """dummy color"""
        return 1

    def init(self):
        """init display"""
        self.driver.init()
        Area.init(self)
        Chip.init(self)
        self.driver.reset()

    def _set_area(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """set area to work on"""
        self.driver.area = {
            'start_x': pos_x1,
            'start_y': pos_y1,
            'end_x': pos_x2,
            'end_y': pos_y2
        }
        self.driver.pointer = (0, 0)



"""SPI communication driver"""
import time
import spidev
import RPi.GPIO  # pylint: disable=I0011,F0401
from gfxlcd.abstract.driver import Driver
RPi.GPIO.setmode(RPi.GPIO.BCM)


class SPI(Driver):
    """SPI communication driver"""
    def __init__(self, spi=0, speed=1000000):
        self.pins = {
            'CS': 8,
            'RST': 25,
            'RS': 24,
            'LED': None
        }
        self.spi = spidev.SpiDev()
        self.spi.open(spi, 0)
        self.spi.max_speed_hz = speed
        self.spi.mode = 0

    def init(self):
        """initialize pins"""
        for pin in self.pins:
            if self.pins[pin] is not None:
                RPi.GPIO.setup(self.pins[pin], RPi.GPIO.OUT)
                RPi.GPIO.output(self.pins[pin], 0)

        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 1)

        if self.pins['LED']:
            RPi.GPIO.output(self.pins['LED'], 1)

    def reset(self):
        """reset a display"""
        if self.pins['LED']:
            RPi.GPIO.output(self.pins['LED'], 1)
        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 1)
        RPi.GPIO.output(self.pins['RST'], 1)
        time.sleep(0.005)
        RPi.GPIO.output(self.pins['RST'], 0)
        time.sleep(0.005)
        RPi.GPIO.output(self.pins['RST'], 1)
        time.sleep(0.005)

    def cmd(self, data, enable):
        """send command to display"""
        RPi.GPIO.output(self.pins['RS'], 0)
        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 0)
        self.spi.xfer2([data])
        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 1)

    def data(self, data, enable):
        """send data to display"""
        RPi.GPIO.output(self.pins['RS'], 1)
        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 0)
        self.spi.xfer2([data >> 8, data])
        if self.pins['CS']:
            RPi.GPIO.output(self.pins['CS'], 1)



import RPi.GPIO
from gfxlcd.abstract.touch import Touch


class AD7843(Touch):
    """AD7843 class"""
    def __init__(self, width, height, int_pin=None, callback=None, cs_pin=None, spi=0, speed=1000000):
        super().__init__(width, height, int_pin, callback, cs_pin, spi, speed)
        self.correction = {
            'x': 364,
            'y': 430,
            'ratio_x': 14.35,
            'ratio_y': 10.59
        }

    def _get_xy(self, offset_x, offset_y):
        """correct x and y"""
        if self.rotate == 0:
            return int((offset_x - self.correction['x']) / self.correction['ratio_x']), \
                int((offset_y - self.correction['y']) / self.correction['ratio_y'])

        if self.rotate == 90:
            return self.height - int((offset_y - self.correction['y']) / self.correction['ratio_y']), \
                int((offset_x - self.correction['x']) / self.correction['ratio_x'])

        if self.rotate == 180:
            return self.width - int((offset_x - self.correction['x']) / self.correction['ratio_x']), \
                self.height - int((offset_y - self.correction['y']) / self.correction['ratio_y'])

        if self.rotate == 270:
            return int((offset_y - self.correction['y']) / self.correction['ratio_y']), \
                self.width - int((offset_x - self.correction['x']) / self.correction['ratio_x'])

    def get_position(self):
        """get touch coords"""
        buffer = []
        fuse = 40
        while len(buffer) < 20 and fuse > 0:
            if self.cs_pin:
                RPi.GPIO.output(self.cs_pin, 0)
            self.spi.xfer2([0xd0])
            recvx = self.spi.readbytes(2)
            self.spi.xfer2([0x90])
            recvy = self.spi.readbytes(2)
            if self.cs_pin:
                RPi.GPIO.output(self.cs_pin, 1)

            tc_rx = recvx[0] << 5
            tc_rx |= recvx[1] >> 3

            tc_ry = recvy[0] << 5
            tc_ry |= recvy[1] >> 3

            pos_x, pos_y = self._get_xy(tc_rx, tc_ry)
            if self._in_bounds(pos_x, pos_y):
                buffer.append((pos_x, pos_y))
            fuse -= 1

        return self._calculate_avr(buffer)



"""Font abstract"""
import abc


class Font(metaclass=abc.ABCMeta):
    font = []  # Dictionary with hex that describe each char
    size = (0, 0)

    def __init__(self):
        pass

    def get(self, letter):
        """return array with letter"""
        return self.font[ord(letter)]

    @property
    def size(self):
        """get font size"""
        return self.size



"""Common drawing functions"""
import math


class Pixel(object):
    """Pixel class"""
    def __init__(self, driver):
        self.driver = driver
        self.rotation = 0
        self.options['color'] = {
            'R': 255, 'G': 255, 'B': 255
        }
        self.options['background_color'] = {
            'R': 0, 'G': 0, 'B': 0,
        }
        self.options['threshold'] = 50
        self.options['transparency_color'] = None

    @property
    def threshold(self):
        """get threshold for B&W conversion"""
        return self.options['threshold']

    @threshold.setter
    def threshold(self, threshold):
        """set B&W threshold for conversion """
        self.options['threshold'] = threshold

    @property
    def transparency_color(self):
        """get transparency color"""
        return self.options['transparency_color']

    @transparency_color.setter
    def transparency_color(self, transparency_color):
        """set transparency color """
        self.options['transparency_color'] = transparency_color

    def draw_pixel(self, pos_x, pos_y):
        """dummy function"""
        pass

    def draw_line(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """dummy function"""
        pass

    def draw_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a rectangle"""
        self.draw_line(pos_x1, pos_y1, pos_x2, pos_y1)
        self.draw_line(pos_x1, pos_y2, pos_x2, pos_y2)
        self.draw_line(pos_x1, pos_y1, pos_x1, pos_y2)
        self.draw_line(pos_x2, pos_y1, pos_x2, pos_y2)

    def draw_circle(self, pos_x, pos_y, radius):
        """draw a circle"""
        err = 0
        offset_x = radius
        offset_y = 0
        while offset_x >= offset_y:
            self.draw_pixel(pos_x + offset_x, pos_y + offset_y)
            self.draw_pixel(pos_x + offset_y, pos_y + offset_x)
            self.draw_pixel(pos_x - offset_y, pos_y + offset_x)
            self.draw_pixel(pos_x - offset_x, pos_y + offset_y)
            self.draw_pixel(pos_x - offset_x, pos_y - offset_y)
            self.draw_pixel(pos_x - offset_y, pos_y - offset_x)
            self.draw_pixel(pos_x + offset_y, pos_y - offset_x)
            self.draw_pixel(pos_x + offset_x, pos_y - offset_y)
            if err <= 0:
                offset_y += 1
                err += 2*offset_y + 1
            else:
                offset_x -= 1
                err -= 2*offset_x + 1

    def draw_arc(self, pos_x, pos_y, radius, start, end):
        """draw an arc"""
        start = start * math.pi / 180
        end = end * math.pi / 180

        err = 0
        offset_x = radius
        offset_y = 0
        while offset_x >= offset_y:
            if start <= math.atan2(offset_y, offset_x) <= end:
                self.draw_pixel(pos_x + offset_x, pos_y + offset_y)
            if start <= math.atan2(offset_x, offset_y) <= end:
                self.draw_pixel(pos_x + offset_y, pos_y + offset_x)
            if start <= math.atan2(offset_x, -offset_y) <= end:
                self.draw_pixel(pos_x - offset_y, pos_y + offset_x)
            if start <= math.atan2(offset_y, -offset_x) <= end:
                self.draw_pixel(pos_x - offset_x, pos_y + offset_y)

            if start <= math.atan2(-offset_y, -offset_x) + 2*math.pi <= end:
                self.draw_pixel(pos_x - offset_x, pos_y - offset_y)
            if start <= math.atan2(-offset_x, -offset_y) + 2*math.pi <= end:
                self.draw_pixel(pos_x - offset_y, pos_y - offset_x)
            if start <= math.atan2(-offset_x, offset_y) + 2*math.pi <= end:
                self.draw_pixel(pos_x + offset_y, pos_y - offset_x)
            if start <= math.atan2(-offset_y, offset_x) + 2*math.pi <= end:
                self.draw_pixel(pos_x + offset_x, pos_y - offset_y)

            if err <= 0:
                offset_y += 1
                err += 2*offset_y + 1
            else:
                offset_x -= 1
                err -= 2*offset_x + 1

    def _calculate_line_steps(self, length, step, required_length):
        """calculate lineparts - helper"""
        steps = [length for _ in range(0, step)]
        if step * length < required_length:
            offset = len(steps) // 2
            rest = required_length - step * length
            steps_even = True if len(steps) & 1 == 0 else False
            rest_even = True if rest & 1 == 0 else False
            appendix = 0
            for idx in range(0, rest):
                steps[offset + appendix] += 1
                if steps_even:
                    appendix = self._calculate_line_appendix(appendix)
                elif idx > 0 and rest_even:
                    appendix = self._calculate_line_appendix(appendix)
                elif not rest_even:
                    appendix = self._calculate_line_appendix(appendix)

        return steps

    def _calculate_line_appendix(self, appendix):
        """calculate appendix during drawing a line"""
        if appendix == 0:
            appendix = -1
        elif appendix < 0:
            appendix *= -1
        else:
            appendix = (appendix + 1) * -1

        return appendix

    def draw_text(self, pos_x, pos_y, text):
        """draw a text"""
        font = self.options['font']
        idx = 0
        for letter in text:
            self._draw_letter(pos_x + idx, pos_y, letter)
            idx += font.size[0]

    def _draw_letter(self, pos_x, pos_y, letter):
        """draw a letter"""
        font = self.options['font']
        bits = font.size[0]
        for row, data in enumerate(font.get(letter)):
            for bit in range(bits):
                if data & 0x01:
                    self.draw_pixel(pos_x + bit, pos_y + row)
                data >>= 1




"""ILI9486 chip driver"""
import time
from gfxlcd.drawing.area import Area
from gfxlcd.abstract.chip import Chip


class ILI9486(Area, Chip):
    """Class for ILI9486 based LCD"""
    rotations = {0: 0x88, 90: 0xf8, 180: 0x48, 270: 0x28}

    def __init__(self, width, height, driver):
        Chip.__init__(self, width, height, driver, True)
        Area.__init__(self, driver)

    def _converted_background_color(self):
        """color from 8-8-8 to 5-6-5"""
        rgb = self.options['background_color']['R'] << 16 | \
            self.options['background_color']['G'] << 8 | \
            self.options['background_color']['B']
        return ((rgb & 0x00f80000) >> 8) |\
            ((rgb & 0x0000fc00) >> 5) | ((rgb & 0x000000f8) >> 3)

    def _converted_color(self):
        """color from 8-8-8 to 5-6-5"""
        rgb = self.options['color']['R'] << 16 | \
            self.options['color']['G'] << 8 | \
            self.options['color']['B']
        return ((rgb & 0x00f80000) >> 8) |\
            ((rgb & 0x0000fc00) >> 5) | ((rgb & 0x000000f8) >> 3)

    def init(self):
        """init display"""
        self.driver.init()
        Area.init(self)
        Chip.init(self)
        self.driver.reset()

        #Read Display MADCTL
        self.driver.cmd(0x0b, None)
        self.driver.data(0x00, None)
        self.driver.data(0x00, None)

        #Sleep OUT
        self.driver.cmd(0x11, None)

        #Interface Pixel Format
        self.driver.cmd(0x3a, None)
        self.driver.data(0x55, None) #0x66 5-6-5 / 55 6-6-6

        #Memory Access Control (
        self.driver.cmd(0x36, None)
        self.driver.data(self.rotations[self.rotation], None)

        #Power Control 3 (For Normal Mode)
        self.driver.cmd(0xc2, None)
        self.driver.data(0x44, None)

        #VCOM Control
        self.driver.cmd(0xc5, None)
        self.driver.data(0x00, None)
        self.driver.data(0x00, None)
        self.driver.data(0x00, None)
        self.driver.data(0x00, None)

        #PGAMCTRL(Positive Gamma Control)
        self.driver.cmd(0xe0, None)
        self.driver.data(0x0F, None)
        self.driver.data(0x1F, None)
        self.driver.data(0x1C, None)
        self.driver.data(0x0C, None)
        self.driver.data(0x0F, None)
        self.driver.data(0x08, None)
        self.driver.data(0x48, None)
        self.driver.data(0x98, None)
        self.driver.data(0x37, None)
        self.driver.data(0x0A, None)
        self.driver.data(0x13, None)
        self.driver.data(0x04, None)
        self.driver.data(0x11, None)
        self.driver.data(0x0D, None)
        self.driver.data(0x00, None)

        #NGAMCTRL (Negative Gamma Correction)
        self.driver.cmd(0xe1, None)
        self.driver.data(0x0F, None)
        self.driver.data(0x32, None)
        self.driver.data(0x2E, None)
        self.driver.data(0x0B, None)
        self.driver.data(0x0D, None)
        self.driver.data(0x05, None)
        self.driver.data(0x47, None)
        self.driver.data(0x75, None)
        self.driver.data(0x37, None)
        self.driver.data(0x06, None)
        self.driver.data(0x10, None)
        self.driver.data(0x03, None)
        self.driver.data(0x24, None)
        self.driver.data(0x20, None)
        self.driver.data(0x00, None)

        #Digital Gamma Control 1
        self.driver.cmd(0xe2, None)
        self.driver.data(0x0F, None)
        self.driver.data(0x32, None)
        self.driver.data(0x2E, None)
        self.driver.data(0x0B, None)
        self.driver.data(0x0D, None)
        self.driver.data(0x05, None)
        self.driver.data(0x47, None)
        self.driver.data(0x75, None)
        self.driver.data(0x37, None)
        self.driver.data(0x06, None)
        self.driver.data(0x10, None)
        self.driver.data(0x03, None)
        self.driver.data(0x24, None)
        self.driver.data(0x20, None)
        self.driver.data(0x00, None)

        #Sleep OUT
        self.driver.cmd(0x11, None)

        #Display ON
        self.driver.cmd(0x29, None)

    def _set_area(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """select area to work with"""
        self.driver.cmd(0x2a, None)
        self.driver.data(pos_x1 >> 8, None)
        self.driver.data(pos_x1 & 0xff, None)
        self.driver.data(pos_x2 >> 8, None)
        self.driver.data(pos_x2 & 0xff, None)
        self.driver.cmd(0x2b, None)
        self.driver.data(pos_y1 >> 8, None)
        self.driver.data(pos_y1 & 0xff, None)
        self.driver.data(pos_y2 >> 8, None)
        self.driver.data(pos_y2 & 0xff, None)
        self.driver.cmd(0x2c, None)



"""Touch panel interface"""
import spidev  # pylint: disable=I0011,F0401
import abc
import RPi.GPIO


class Touch(metaclass=abc.ABCMeta):
    """Touch class"""
    def __init__(self, width, height, int_pin=None, callback=None, cs_pin=None, spi=0, speed=1000000):
        self.width = width
        self.height = height
        self.spi = spidev.SpiDev()
        self.spi.open(spi, 0)
        self.spi.max_speed_hz = speed
        self.spi.mode = 0
        self.cs_pin = cs_pin
        self.int_pin = int_pin
        self.callback = callback
        self.bouncetime = 500
        self.rotate = 0
        self.correction = {
            'x': 0,
            'y': 0,
            'ratio_x': 1,
            'ratio_y': 1,
        }

    def init(self):
        """some init functions"""
        if self.int_pin:
            RPi.GPIO.setup(self.int_pin, RPi.GPIO.IN)
            RPi.GPIO.add_event_detect(
                self.int_pin, RPi.GPIO.BOTH, callback=self._interrupt, bouncetime=self.bouncetime
            )
        if self.cs_pin:
            RPi.GPIO.setup(self.cs_pin, RPi.GPIO.OUT)
            RPi.GPIO.output(self.cs_pin, 1)

    @abc.abstractmethod
    def get_position(self):
        """returns pressed position"""
        return

    def close(self):
        """close action"""
        if self.int_pin:
            RPi.GPIO.remove_event_detect(self.int_pin)
        self.spi.close()

    def _interrupt(self, channel):
        """call users callback"""
        self.callback(self.get_position())

    def _calculate_avr(self, points):
        """calculate x,y by average"""
        if len(points) == 0:
            return None
        sum_x = 0
        sum_y = 0
        for point in points:
            sum_x += point[0]
            sum_y += point[1]

        return int(sum_x / len(points)), int(sum_y / len(points))

    def _in_bounds(self, pos_x, pos_y):
        """checks if point is in range"""
        if self.rotate == 0 or self.rotate == 180:
            return 0 <= pos_x <= self.width and 0 <= pos_y <= self.height
        else:
            return 0 <= pos_y <= self.width and 0 <= pos_x <= self.height



"""Area driver """
import time
from gfxlcd.abstract.driver import Driver


class AreaDriver(Driver):
    """Null communication driver"""
    def __init__(self, width, height):
        self.height = height
        self.width = width
        self.buffer = [[0] * self.height for x in range(self.width)]
        self.area = {
            'start_x': 0,
            'start_y': 0,
            'end_x': width,
            'end_y': height
        }
        self.pointer = (0, 0)

    def init(self):
        """initialize pins"""
        pass

    def reset(self):
        """reset a display"""
        pass

    def cmd(self, data, enable):
        """send command to display"""
        pass

    def data(self, data, enable):
        """send data to display"""
        app_x, app_y = self.pointer
        self.buffer[self.area['start_x'] + app_x][self.area['start_y'] + app_y] = data
        self._inc_pointer()

    def _inc_pointer(self):
        app_x, app_y = self.pointer
        app_x += 1
        if self.area['start_x'] + app_x > self.area['end_x']:
            app_x = 0
            app_y += 1

        if self.area['start_y'] + app_y > self.area['end_y']:
            app_x = 0
            app_y = 0

        self.pointer = (app_x, app_y)




"""Null Page test chip driver"""
from gfxlcd.drawing.page import Page
from gfxlcd.abstract.chip import Chip


class NullPage(Page, Chip):
    """Test chip driver for page drawing"""
    def __init__(self, width, height, driver, auto_flush=True):
        Chip.__init__(self, width, height, driver, auto_flush)
        Page.__init__(self, driver)
        self.rotation = 0
        self.buffer = []
        self.area = {
            'start_x': 0,
            'start_y': 0,
            'end_x': width-1,
            'end_y': height-1
        }

    def init(self):
        """inits a device"""
        Page.init(self)
        Chip.init(self)

    def _converted_background_color(self):
        """convert RGB background to available color"""
        return 1

    def _converted_color(self):
        """convert RGB color to available color"""
        return 1

    def flush(self, force=None):
        """flush buffer to device
        :force - boolean|None"""
        return self.buffer

    def set_area(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """set area to work on"""
        self.area = {
            'start_x': pos_x1,
            'start_y': pos_y1,
            'end_x': pos_x2,
            'end_y': pos_y2
        }
